from __future__ import absolute_import
import contextlib
import enum
import tornado.ioloop
from tornado import gen
import tornado.gen
from tornado.gen import coroutine, Return

from . import exceptions
from . import tools


@enum.unique
class ConfirmationTypes(enum.Enum):
    ACK = 'ack'
    NACK = 'nack'


def ensure_coroutine(func_or_coro):
    if gen.is_coroutine_function(func_or_coro):
        return func_or_coro
    else:

        @coroutine
        def coro(*args, **kwargs):
            raise Return(func_or_coro(*args, **kwargs))

        return coro


class _CallbackWrapper(object):
    """
    Take a callback that expects the first argument to be the object corresponding
    to the source of the message and replace it with a proxy object that the
    function will receive as the source object instead.  This wrapper will
    also correctly call the callback in the correct way whether it is a coroutine
    or plain function.
    """

    def __init__(self, source_proxy, callback):
        self._proxy = source_proxy
        self._callback = ensure_coroutine(callback)

    @gen.coroutine
    def __call__(self, unused_source, *args, **kwargs):
        result = yield self._callback(self._proxy, *args, **kwargs)
        raise Return(result)


def future_with_timeout(loop, timeout, future=None):
    """
    Create a future with a timeout

    :param loop: The tornado event loop
    :param timeout: The timeout in seconds
    :param future: An optional existing future
    :return:
    """
    loop = loop if loop else tornado.ioloop.IOLoop.current()
    f = future or tools.create_future(loop=loop)

    def on_timeout():
        if f.done():
            return
        f.set_exception(tornado.ioloop.TimeoutError)

    if timeout:
        handle = loop.call_later(timeout, on_timeout)

        def on_result(_unused_future):
            # Cancel the timeout if the future is done
            loop.remove_timeout(handle)

        f.add_done_callback(on_result)

    return f


class FutureStore(object):
    """
    Borrowed from aio_pika (https://github.com/mosquito/aio-pika)
    """
    __slots__ = "__collection", "__loop", "__parent_store"

    def __init__(self, loop, parent_store=None):
        self.__parent_store = parent_store
        self.__collection = set()
        self.__loop = loop if loop else tornado.ioloop.IOLoop.current()

    def _on_future_done(self, future):
        if future in self.__collection:
            self.__collection.remove(future)

    @staticmethod
    def _reject_future(future, exception):
        if future.done():
            return

        future.set_exception(exception)

    def add(self, future):
        if self.__parent_store:
            self.__parent_store.add(future)

        self.__collection.add(future)
        future.add_done_callback(self._on_future_done)

    def reject_all(self, exception):
        for future in list(self.__collection):
            self.__collection.remove(future)
            self.__loop.add_callback(self._reject_future, future, exception)

    @staticmethod
    def _on_timeout(future):
        if future.done():
            return

        future.set_exception(tornado.ioloop.TimeoutError)

    def create_future(self, timeout=None):
        future = future_with_timeout(self.__loop, timeout)

        self.add(future)

        if self.__parent_store:
            self.__parent_store.add(future)

        return future

    def create_child(self):
        return FutureStore(self.__loop, parent_store=self)

    @contextlib.contextmanager
    def pending_future(self, timeout=None):
        future = None
        try:
            future = self.create_future(timeout)
            yield future
        finally:
            # Cleanup
            if future in self.__collection:
                self.__collection.remove(future)


class BaseChannel(object):
    __slots__ = ('_channel_futures', 'loop', '_futures', '_closing')

    def __init__(self, loop, future_store):
        """
        :type loop: :class:`tornado.ioloop.IOLoop`
        :type future_store: :class:`FutureStore`
        """
        self.loop = loop
        self._futures = future_store
        self._closing = tools.create_future(loop=self.loop)

    @property
    def is_closed(self):
        return self._closing.done()

    def _create_future(self, timeout=None):
        f = self._futures.create_future(timeout)
        return f

    @staticmethod
    def _ensure_channel_is_open(func):

        @contextlib.wraps(func)
        @tools.coroutine
        def wrap(self, *args, **kwargs):
            if self.is_closed:
                raise RuntimeError("The channel is closed")

            raise gen.Return((yield func(self, *args, **kwargs)))

        return wrap

    def __repr__(self):
        return "<{}: {}>".format(self.__class__.__name__, getattr(self, 'name', id(self)))



from __future__ import absolute_import

# Disable variable name conventions because otherwise all the variables
# in this file would have to be upper case
# pylint: disable=invalid-name

author_info = (('Martin Uhrin', 'martin.uhrin@gmail.com'),)

package_info = "A backport of aio-pika for Tornado to support python 2.7+."
package_license = "Apache Software License"

version_info = (0, 2, 1)

__author__ = ", ".join("{} <{}>".format(*info) for info in author_info)
__version__ = ".".join(tuple(map(str, version_info)))



from __future__ import absolute_import
from collections import namedtuple
import contextlib
from logging import getLogger

import pika.spec
from tornado import gen, locks

from .exchange import Exchange
from .message import IncomingMessage
from .common import BaseChannel
from . import tools
from .exceptions import QueueEmpty

LOGGER = getLogger(__name__)

ConsumerTag = str
DeclarationResult = namedtuple('DeclarationResult', ('message_count', 'consumer_count'))


class Queue(BaseChannel):
    """ AMQP queue abstraction """

    __slots__ = ('name', 'durable', 'exclusive', 'auto_delete', 'arguments', '_get_lock', '_channel', '__closing',
                 'declaration_result')

    def __init__(self, loop, future_store, channel, name, durable, exclusive, auto_delete, arguments):  # pylint: disable=too-many-arguments
        """
        :type loop: :class:`tornado.ioloop.IOLoop`
        :type future_store: :class:`topika.FutureStore`
        :type channel: :class:`pika.Channel`
        :type name: str
        :type durable: bool
        :type exclusive: bool
        :type auto_delete: bool
        :type arguments: dict
        """
        super(Queue, self).__init__(loop, future_store)

        self._channel = channel
        self.name = name or ''
        self.durable = durable
        self.exclusive = exclusive
        self.auto_delete = auto_delete
        self.arguments = arguments
        self.declaration_result = None  # type: DeclarationResult
        self._get_lock = locks.Lock()

    def __str__(self):
        return "%s" % self.name

    def __repr__(self):
        return "<Queue(%s): auto_delete=%s, durable=%s, exclusive=%s, arguments=%r>" % (
            self,
            self.auto_delete,
            self.durable,
            self.exclusive,
            self.arguments,
        )

    @BaseChannel._ensure_channel_is_open
    def declare(self, timeout=None, passive=False):
        """ Declare queue.

        :param timeout: execution timeout
        :type timeout: int
        :param passive: Only check to see if the queue exists.
        :type passive: bool
        :rtype: :class:`tornado.concurrent.Future`
        """

        LOGGER.debug("Declaring queue: %r", self)

        declare_future = self._create_future(timeout)

        self._channel.queue_declare(
            queue=self.name,
            passive=passive,
            durable=self.durable,
            exclusive=self.exclusive,
            auto_delete=self.auto_delete,
            arguments=self.arguments,
            callback=declare_future.set_result)

        def on_queue_declared(result):
            res = result.result()
            self.name = res.method.queue
            self.declaration_result = DeclarationResult(
                message_count=res.method.message_count,
                consumer_count=res.method.consumer_count,
            )

        declare_future.add_done_callback(on_queue_declared)

        return declare_future

    @BaseChannel._ensure_channel_is_open
    def bind(self, exchange, routing_key=None, arguments=None, timeout=None):
        """ A binding is a relationship between an exchange and a queue. This can be
        simply read as: the queue is interested in messages from this exchange.

        Bindings can take an extra routing_key parameter. To avoid the confusion
        with a basic_publish parameter we're going to call it a binding key.

        :param exchange: :class:`topika.exchange.Exchange` instance
        :type exchange: :class:`ExchangeType`
        :param routing_key: routing key
        :type routing_key: str
        :param arguments: additional arguments (will be passed to `pika`)
        :type arguments: dict or NoneType
        :param timeout: execution timeout
        :type timeout: int
        :raises tornado.gen.TimeoutError: when the binding timeout period has elapsed.
        :rtype: :class:`tornado.concurrent.Future`
        """
        LOGGER.debug("Binding queue %r: exchange=%r, routing_key=%r, arguments=%r", self, exchange, routing_key,
                     arguments)

        bind_future = self._create_future(timeout)

        self._channel.queue_bind(
            queue=self.name,
            exchange=Exchange._get_exchange_name(exchange),  # pylint: disable=protected-access
            routing_key=routing_key,
            arguments=arguments,
            callback=bind_future.set_result)

        return bind_future

    @BaseChannel._ensure_channel_is_open
    def unbind(self, exchange, routing_key, arguments=None, timeout=None):
        """ Remove binding from exchange for this :class:`Queue` instance

        :param exchange: :class:`topika.exchange.Exchange` instance
        :type exchange: :class:`ExchangeType`
        :param routing_key: routing key
        :type routing_key: str
        :param arguments: additional arguments (will be passed to `pika`)
        :type arguments: dict or NoneType
        :param timeout: execution timeout
        :type timeout: int
        :raises tornado.gen.TimeoutError: when the unbinding timeout period has elapsed.
        :rtype: :class:`tornado.concurrent.Future`
        """

        LOGGER.debug("Unbinding queue %r: exchange=%r, routing_key=%r, arguments=%r", self, exchange, routing_key,
                     arguments)

        unbind_future = self._create_future(timeout)

        self._channel.queue_unbind(
            queue=self.name,
            exchange=Exchange._get_exchange_name(exchange),  # pylint: disable=protected-access
            routing_key=routing_key,
            arguments=arguments,
            callback=unbind_future.set_result)

        return unbind_future

    @BaseChannel._ensure_channel_is_open
    @tools.coroutine
    def consume(self, callback, no_ack=False, exclusive=False, arguments=None, consumer_tag=None, timeout=None):  # pylint: disable=too-many-arguments
        """ Start to consuming the :class:`Queue`.

        :param timeout: :class:`tornado.gen.TimeoutError` will be raises when the
                        Future was not finished after this time.
        :param callback: Consuming callback. Could be a coroutine.
        :type callback: :class:`FunctionType`
        :param no_ack: if :class:`True` you don't need to call :func:`topika.message.IncomingMessage.ack`
        :type no_ack: bool
        :param exclusive: Makes this queue exclusive. Exclusive queues may only be accessed by the current
                          connection, and are deleted when that connection closes. Passive declaration of an
                          exclusive queue by other connections are not allowed.
        :type exclusive: bool
        :param arguments: extended arguments for pika
        :type arguments: Optional[dict]
        :param consumer_tag: optional consumer tag

        :raises tornado.gen.TimeoutError: when the consuming timeout period has elapsed.
        :rtype: class:`Generator[Any, None, ConsumerTag]`
        """

        LOGGER.debug("Start to consuming queue: %r", self)
        future = self._futures.create_future(timeout=timeout)

        def consumer(channel, envelope, properties, body):
            """
            :type channel: :class:`Channel`
            :type body: bytes
            """
            message = IncomingMessage(
                channel=channel,
                body=body,
                envelope=envelope,
                properties=properties,
                no_ack=no_ack,
            )

            if tools.iscoroutinepartial(callback):
                tools.create_task(callback(message))
            else:
                self.loop.add_callback(callback, message)

        consumer_tag = self._channel.basic_consume(
            queue=self.name,
            on_message_callback=consumer,
            auto_ack=no_ack,
            exclusive=exclusive,
            consumer_tag=consumer_tag,
            arguments=arguments,
            callback=future.set_result,
        )

        yield future

        raise gen.Return(consumer_tag)

    @BaseChannel._ensure_channel_is_open
    def cancel(self, consumer_tag, timeout=None):
        """ This method cancels a consumer. This does not affect already
        delivered messages, but it does mean the server will not send any more
        messages for that consumer. The client may receive an arbitrary number
        of messages in between sending the cancel method and receiving the
        cancel-ok reply. It may also be sent from the server to the client in
        the event of the consumer being unexpectedly cancelled (i.e. cancelled
        for any reason other than the server receiving the corresponding
        basic.cancel from the client). This allows clients to be notified of
        the loss of consumers due to events such as queue deletion.

        :param consumer_tag: consumer tag returned by :func:`~topika.Queue.consume`
        :type consumer_tag: :class:`ConsumerTag`
        :param timeout: execution timeout
        :type timeout: int or NoneType
        :return: Basic.CancelOk when operation completed successfully
        """
        cancel_future = self._create_future(timeout)
        self._channel.basic_cancel(consumer_tag=consumer_tag, callback=cancel_future.set_result)

        return cancel_future

    @BaseChannel._ensure_channel_is_open
    @gen.coroutine
    def get(self, no_ack=False, timeout=None, fail=True):
        """ Get message from the queue.

        :param no_ack: if :class:`True` you don't need to call
                       :func:`aio_pika.message.IncomingMessage.ack`
        :type no_ack: bool
        :param timeout: execution timeout
        :type timeout: int or NoneType
        :param fail: Should return :class:`None` instead of raise an
                     exception :class:`aio_pika.exceptions.QueueEmpty`.
        :type fail: bool
        :rtype: :class:`Generator[Any, None, Optional[IncomingMessage]]`
        """

        get_future = self._create_future(timeout)

        def _on_getempty(method_frame, *_a, **_kw):
            if fail:
                get_future.set_exception(QueueEmpty(method_frame))
            else:
                get_future.set_result(None)

        def _on_getok(channel, envelope, props, body):
            message = IncomingMessage(
                channel,
                envelope,
                props,
                body,
                no_ack=no_ack,
            )

            get_future.set_result(message)

        with (yield self._get_lock.acquire()), self._capture_empty(_on_getempty):
            LOGGER.debug("Awaiting message from queue: %r", self)

            self._channel.basic_get(queue=self.name, callback=_on_getok, auto_ack=no_ack)

            try:
                message = yield get_future
                raise gen.Return(message)
            finally:
                self._channel._on_getempty = None  # pylint: disable=protected-access

    @contextlib.contextmanager
    def _capture_empty(self, callback):
        """
        Capture an empty message from the queue.
        :param callback: The callback to call upon receiving the empty message
        """
        cancelled = False

        def callback_wrapper(frame):
            if not cancelled:
                callback(frame)

        self._channel.add_callback(callback_wrapper, (pika.spec.Basic.GetEmpty,), one_shot=True)
        try:
            yield
        finally:
            cancelled = True

    @BaseChannel._ensure_channel_is_open
    def purge(self, timeout=None):
        """ Purge all messages from the queue.

        :param timeout: execution timeout
        :rtype: :class:`tornado.concurrent.Future`
        """

        LOGGER.info("Purging queue: %r", self)

        purge_future = self._create_future(timeout)
        self._channel.queue_purge(queue=self.name, callback=purge_future.set_result)
        return purge_future

    @BaseChannel._ensure_channel_is_open
    def delete(self, if_unused=True, if_empty=True, timeout=None):
        """ Delete the queue.

        :param if_unused: Perform delete only when unused
        :param if_empty: Perform delete only when empty
        :param timeout: execution timeout
        :rtype: :class:`tornado.concurrent.Future`
        """

        LOGGER.info("Deleting %r", self)

        self._futures.reject_all(RuntimeError("Queue was deleted"))

        future = self._create_future(timeout)

        self._channel.queue_delete(queue=self.name, if_unused=if_unused, if_empty=if_empty, callback=future.set_result)

        return future

    # def __iter__(self) -> 'QueueIterator':
    #     """ Return the :class:`QueueIterator` which might be used with `async for` syntax
    #     before use it we are strongly recommended call :method:`set_qos` with argument `1`. """
    #     iterator = self.iterator()
    #     self.loop.add_callback(iterator.consume)
    #     return iterator
    #
    # @asyncio.coroutine
    # def __aiter__(self) -> 'QueueIterator':
    #     iterator = self.iterator()
    #     yield from iterator.consume()
    #     return iterator
    #
    # def iterator(self) -> 'QueueIterator':
    #     """ Returns an iterator for async for expression.
    #
    #     Full example:
    #
    #     .. code-block:: python
    #
    #         import aio_pika
    #
    #         async def main():
    #             connection = await aio_pika.connect()
    #
    #             async with connection:
    #                 channel = await connection.channel()
    #
    #                 queue = await channel.declare_queue('test')
    #
    #                 async with queue.iterator() as q:
    #                     async for message in q:
    #                         print(message.body)
    #
    #     When your program runs with run_forever the iterator will be closed
    #     in background. In this case the context processor for iterator might
    #     be skipped and the queue might be used in the "async for"
    #     expression directly.
    #
    #     .. code-block:: python
    #
    #         import aio_pika
    #
    #         async def main():
    #             connection = await aio_pika.connect()
    #
    #             async with connection:
    #                 channel = await connection.channel()
    #
    #                 queue = await channel.declare_queue('test')
    #
    #                 async for message in queue:
    #                     print(message.body)
    #
    #     :return: QueueIterator
    #     """
    #
    #     return QueueIterator(self)


#
# class QueueIterator:
#     def __init__(self, queue: Queue):
#         self._amqp_queue = queue
#         self._queue = asyncio.Queue(loop=self.loop)
#         self._consumer_tag = None
#
#     @property
#     def loop(self) -> asyncio.AbstractEventLoop:
#         return self._amqp_queue.loop
#
#     def on_message(self, message: IncomingMessage):
#         self._queue.put_nowait(message)
#
#     @asyncio.coroutine
#     def consume(self):
#         self._consumer_tag = yield from self._amqp_queue.consume(self.on_message)
#
#     @asyncio.coroutine
#     def _close(self):
#         yield from self._amqp_queue.cancel(self._consumer_tag)
#         self._consumer_tag = None
#
#         def get_msg():
#             try:
#                 return self._queue.get_nowait()
#             except asyncio.QueueEmpty:
#                 return
#
#         # Reject all messages
#         msg = get_msg()     # type: IncomingMessage
#         while msg:
#             msg.reject(requeue=True)
#             msg = get_msg()  # type: IncomingMessage
#
#     def close(self) -> asyncio.Future:
#         if not self._consumer_tag or self._amqp_queue._channel.is_closed:
#             f = asyncio.Future(loop=self.loop)
#             f.set_result(None)
#             return f
#
#         return self.loop.add_callback(self._close)
#
#     def __del__(self):
#         self.close()
#
#     def __iter__(self):
#         return self
#
#     __aiter__ = asyncio.coroutine(__iter__)
#
#     @asyncio.coroutine
#     def __next__(self) -> IncomingMessage:
#         try:
#             return (yield from self._queue.get())
#         except asyncio.CancelledError:
#             yield from self.close()
#             raise
#
#     @asyncio.coroutine
#     def __aenter__(self):
#         if self._consumer_tag is None:
#             yield from self.consume()
#         return self
#
#     @asyncio.coroutine
#     def __aexit__(self, exc_type, exc_val, exc_tb):
#         yield from self.close()
#
#     __anext__ = __next__

__all__ = 'Queue'



from __future__ import absolute_import
from logging import getLogger
from typing import Callable, Any, Generator, Union
from tornado import gen

from . import compat
from . import exchange
from . import tools
from .compat import Awaitable
from .exchange import Exchange, ExchangeType
from .message import IncomingMessage
from .queue import Queue
from .common import BaseChannel, FutureStore
from .channel import Channel
from .robust_queue import RobustQueue
from .robust_exchange import RobustExchange

log = getLogger(__name__)

FunctionOrCoroutine = Union[Callable[[IncomingMessage], Any], Awaitable[IncomingMessage]]


class RobustChannel(Channel):
    """ Channel abstraction """

    QUEUE_CLASS = RobustQueue
    EXCHANGE_CLASS = RobustExchange

    def __init__(self,
                 connection,
                 loop,
                 future_store,
                 channel_number=None,
                 publisher_confirms=True,
                 on_return_raises=False):
        """

        :param connection: :class:`pika.adapters.tornado_connection.TornadoConnection` instance
        :param loop: Event loop (:func:`tornado.ioloop.IOLoop.current()` when :class:`None`)
        :param future_store: :class:`topika.common.FutureStore` instance
        :param publisher_confirms: False if you don't need delivery confirmations (in pursuit of performance)
        """
        super(RobustChannel, self).__init__(
            loop=loop,
            future_store=future_store.create_child(),
            connection=connection,
            channel_number=channel_number,
            publisher_confirms=publisher_confirms,
            on_return_raises=on_return_raises,
        )

        self._closed = False
        self._exchanges = dict()
        self._queues = dict()
        self._qos = 0, 0

    @gen.coroutine
    def on_reconnect(self, connection, channel_number):
        exc = compat.ConnectionError('Auto Reconnect Error')

        if not self._closing.done():
            self._closing.set_exception(exc)

        self._closing = tools.create_future(loop=self.loop)
        self._futures.reject_all(exc)
        self._connection = connection
        self._channel_number = channel_number

        yield self.initialize()

        for exchange in self._exchanges.values():
            yield exchange.on_reconnect(self)

        for queue in self._queues.values():
            yield queue.on_reconnect(self)

    @gen.coroutine
    def initialize(self, timeout=None):
        result = yield super(RobustChannel, self).initialize()

        prefetch_count, prefetch_size = self._qos

        yield self.set_qos(prefetch_count=prefetch_count, prefetch_size=prefetch_size)

        raise gen.Return(result)

    @gen.coroutine
    def set_qos(self, prefetch_count=0, prefetch_size=0, global_qos=False, timeout=None):
        if global_qos:
            raise NotImplementedError("Not available to RobustConnection")

        self._qos = prefetch_count, prefetch_size

        raise gen.Return((yield super(RobustChannel, self).set_qos(
            prefetch_count=prefetch_count,
            prefetch_size=prefetch_size,
            timeout=timeout,
        )))

    @BaseChannel._ensure_channel_is_open
    @gen.coroutine
    def close(self):
        if self._closed:
            return

        with (yield self._write_lock.acquire()):
            self._closed = True
            self._channel.close()
            yield self.closing
            self._channel = None

    @gen.coroutine
    def declare_exchange(self,
                         name,
                         type=exchange.ExchangeType.DIRECT,
                         durable=None,
                         auto_delete=False,
                         internal=False,
                         passive=False,
                         arguments=None,
                         timeout=None,
                         robust=True):

        exchange = yield super(RobustChannel, self).declare_exchange(
            name=name,
            type=type,
            durable=durable,
            auto_delete=auto_delete,
            internal=internal,
            passive=passive,
            arguments=arguments,
            timeout=timeout,
        )

        if not internal and robust:
            self._exchanges[name] = exchange

        raise gen.Return(exchange)

    @gen.coroutine
    def exchange_delete(self, exchange_name, timeout=None, if_unused=False):
        result = yield super(RobustChannel, self).exchange_delete(
            exchange_name=exchange_name, timeout=timeout, if_unused=if_unused)

        self._exchanges.pop(exchange_name, None)

        raise gen.Return(result)

    @gen.coroutine
    def declare_queue(self,
                      name=None,
                      durable=None,
                      exclusive=False,
                      passive=False,
                      auto_delete=False,
                      arguments=None,
                      timeout=None,
                      robust=True):
        """
        :param name: queue name
        :type name: str
        :param durable: Durability (queue survive broker restart)
        :type durable: bool
        :param exclusive: Makes this queue exclusive. Exclusive queues may only be \
        accessed by the current connection, and are deleted when that connection \
        closes. Passive declaration of an exclusive queue by other connections are not allowed.
        :type exclusive: bool
        :param passive: Only check to see if the queue exists.
        :type passive: bool
        :param auto_delete: Delete queue when channel will be closed.
        :type auto_delete: bool
        :param arguments: pika additional arguments
        :type arguments: Optional[dict]
        :param timeout: execution timeout
        :type timeout: int
        :type robust: bool
        :rtype: :class:`Generator[Any, None, Queue]`
        """

        queue = yield super(RobustChannel, self).declare_queue(
            name=name,
            durable=durable,
            exclusive=exclusive,
            passive=passive,
            auto_delete=auto_delete,
            arguments=arguments,
            timeout=timeout,
        )

        if robust:
            self._queues[name] = queue

        raise gen.Return(queue)

    @gen.coroutine
    def queue_delete(self, queue_name, timeout=None, if_unused=False, if_empty=False):
        result = yield super(RobustChannel, self).queue_delete(
            queue_name=queue_name, timeout=timeout, if_unused=if_unused, if_empty=if_empty)

        self._queues.pop(queue_name, None)
        raise gen.Return(result)


__all__ = ('RobustChannel',)



from __future__ import absolute_import
from tornado import gen, concurrent

from .exceptions import TransactionClosed


class Transaction:

    def __init__(self, channel, future_store):
        """
        :type future_store: :class:`FutureStore`
        """
        self._channel = channel
        self._future_store = future_store
        self.closing = self._future_store.create_future()  # type: concurrent.Future

    def _create_future(self, timeout=None):
        if self.closing.done():
            raise RuntimeError("Can't reuse closed transaction")

        return self._future_store.create_future(timeout)

    def select(self, timeout=None):
        future = self._create_future(timeout)
        self._channel.tx_select(future.set_result)
        return future

    def rollback(self, timeout=None):
        future = self._create_future(timeout)
        self._channel.tx_rollback(future.set_result)
        return future

    def commit(self, timeout=None):
        future = self._create_future(timeout)
        self._channel.tx_commit(future.set_result)
        return future

    def close(self, exc=TransactionClosed):
        """
        :type exc: Exception
        """
        if not self.closing.done():
            self.closing.set_result(None)
        self._future_store.reject_all(exc)

    @gen.coroutine
    def __aenter__(self):
        """ Only for python 3.5+ """
        result = yield self.select()
        raise gen.Return(result)

    @gen.coroutine
    def __aexit__(self, exc_type, exc_val, exc_tb):
        """ Only for python 3.5+ """
        if exc_type:
            yield self.rollback()
        else:
            yield self.commit()

        self.close()

    def __del__(self):
        self.close(ReferenceError('Transaction deleted'))

    def on_close_callback(self, result):
        """
        :type result: :class:`concurrent.Future`
        """
        exc = result.exception()

        if exc:
            self.close(exc)



from __future__ import absolute_import

from .connection import Connection, connect
from .channel import Channel
from .exchange import Exchange, ExchangeType
from .message import Message, IncomingMessage, DeliveryMode
from .queue import Queue
from .robust_connection import connect_robust
from .exceptions import AMQPException, MessageProcessError
from .compat import ConnectionError, ConnectionRefusedError

from .version import __author__, __version__, author_info, package_info, package_license, version_info

__all__ = ('__author__', '__version__', 'connect', 'connect_robust', 'Connection', 'Channel', 'DeliveryMode',
           'Exchange', 'ExchangeType', 'Queue', 'Message', 'IncomingMessage', 'author_info', 'package_info',
           'version_info', 'package_license', 'AMQPException', 'MessageProcessError', 'ConnectionError',
           'ConnectionRefusedError')



from __future__ import absolute_import
from logging import getLogger
from tornado import gen

from . import compat
from .common import FutureStore
from .exchange import Exchange, ExchangeType
from .channel import Channel

_LOGGER = getLogger(__name__)


class RobustExchange(Exchange):
    """ Exchange abstraction """

    def __init__(self,
                 loop,
                 future_store,
                 channel,
                 publish_method,
                 name=None,
                 type=ExchangeType.DIRECT,
                 passive=False,
                 durable=False,
                 auto_delete=False,
                 internal=False,
                 arguments=None):
        """
        :type loop: :class:`tornado.ioloop.IOLoop`
        :type future_store: :class:`FutureStore`
        :type channel: :class:`Channel`
        :type name: str
        :type type: ExchangeType
        :type auto_delete: Optional[bool]
        :type durable: Optional[bool]
        :type internal: Optional[bool]
        :type passive: Optional[bool]
        :type arguments: dict or NoneType
        """
        super(RobustExchange, self).__init__(
            loop=loop,
            future_store=future_store,
            channel=channel,
            publish_method=publish_method,
            name=name,
            type=type,
            auto_delete=auto_delete,
            durable=durable,
            internal=internal,
            passive=passive,
            arguments=arguments,
        )

        self._bindings = dict()

    @gen.coroutine
    def on_reconnect(self, channel):
        """
        :type channel: :class:`Channel`
        """
        self._futures.reject_all(compat.ConnectionError("Auto Reconnect Error"))
        self._channel = channel._channel

        yield self.declare()

        for exchange, kwargs in self._bindings.items():
            yield self.bind(exchange, **kwargs)

    @gen.coroutine
    def bind(self, exchange, routing_key='', arguments=None, timeout=None):
        """
        :param exchange: :class:`topika.exchange.Exchange` instance
        :type exchange: ExchangeType_
        :param routing_key: routing key
        :type routing_key: str
        :param arguments: additional arguments (will be passed to `pika`)
        :param timeout: execution timeout
        :type timeout: int
        :rtype: :class:`tornado.concurrent.Future`
        """
        result = yield super(RobustExchange, self).bind(
            exchange, routing_key=routing_key, arguments=arguments, timeout=timeout)

        self._bindings[exchange] = dict(routing_key=routing_key, arguments=arguments)

        raise gen.Return(result)

    @gen.coroutine
    def unbind(self, exchange, routing_key='', arguments=None, timeout=None):
        """ Remove exchange-to-exchange binding for this :class:`Exchange` instance

        :param exchange: :class:`topika.exchange.Exchange` instance
        :type exchange: ExchangeType_
        :param routing_key: routing key
        :type routing_key: str
        :param arguments: additional arguments (will be passed to `pika`)
        :type arguments: dict
        :param timeout: execution timeout
        :type timeout: int
        :rtype: :class:`tornado.concurrent.Future`
        """
        result = yield super(RobustExchange, self).unbind(exchange, routing_key, arguments=arguments, timeout=timeout)
        self._bindings.pop(exchange, None)
        raise gen.Return(result)


__all__ = ('Exchange',)



from __future__ import absolute_import
import contextlib

try:
    from typing import Awaitable
except ImportError:
    import six
    from typing import Generic, TypeVar
    from abc import ABCMeta, abstractmethod

    T_co = TypeVar('T_co', covariant=True)

    @six.add_metaclass(ABCMeta)
    class _Awaitable(object):

        __slots__ = ()

        @abstractmethod
        def __await__(self):
            yield

    class Awaitable(Generic[T_co], _Awaitable):
        __slots__ = ()


try:
    from contextlib import suppress
except ImportError:

    @contextlib.contextmanager
    def suppress(*exceptions):
        """
        An attempt to emulate python3's contextlib.suppress
        :param exceptions: The list of exceptions to suppress, all if none are supplied
        """
        exc = exceptions or Exception
        try:
            yield
        except exc:
            pass


try:
    # Either take the native ones
    ConnectionError = ConnectionError
    ConnectionRefusedError = ConnectionRefusedError
except NameError:
    # Or define these ourselves (python2)
    class ConnectionError(Exception):
        pass

    class ConnectionRefusedError(ConnectionError):
        pass


try:
    from contextlib import suppress
except ImportError:

    @contextlib.contextmanager
    def suppress(*exceptions):
        excs = exceptions or Exception
        try:
            yield
        except excs:
            pass


__all__ = ('ConnectionError', 'ConnectionRefusedError')



from __future__ import absolute_import
from collections import namedtuple

from tornado import gen
import shortuuid

from . import compat
from .queue import Queue
from . import tools

DeclarationResult = namedtuple('DeclarationResult', ('message_count', 'consumer_count'))


class RobustQueue(Queue):
    """A queue that, if the connection drops, will recreate itself once it's back up"""

    def __init__(self, loop, future_store, channel, name, durable, exclusive, auto_delete, arguments):
        # pylint: disable=too-many-arguments
        """
        :type loop: :class:`tornado.ioloop.IOLoop`
        :type future_store: :class:`topika.FutureStore`
        :type channel: :class:`pika.Channel`
        :type name: str
        :type durable: bool
        :type exclusive: bool
        :type auto_delete: bool
        :type arguments: dict
        """

        super(RobustQueue, self).__init__(loop, future_store, channel, name or "amq_%s" % shortuuid.uuid(), durable,
                                          exclusive, auto_delete, arguments)

        self._consumers = {}
        self._bindings = {}

    @gen.coroutine
    def on_reconnect(self, channel):
        """
        :type channel: :class:`topika.Channel`
        """
        self._futures.reject_all(compat.ConnectionError("Auto Reconnect Error"))
        self._channel = channel._channel  # pylint: disable=protected-access

        yield self.declare()

        for item, kwargs in self._bindings.items():
            exchange, routing_key = item
            yield self.bind(exchange, routing_key, **kwargs)

        for consumer_tag, kwargs in tuple(self._consumers.items()):
            yield self.consume(consumer_tag=consumer_tag, **kwargs)

    @gen.coroutine
    def bind(self, exchange, routing_key=None, arguments=None, timeout=None):
        """ A binding is a relationship between an exchange and a queue. This can be
        simply read as: the queue is interested in messages from this exchange.

        Bindings can take an extra routing_key parameter. To avoid the confusion
        with a basic_publish parameter we're going to call it a binding key.

        :param exchange: :class:`topika.exchange.Exchange` instance
        :type exchange: :class:`ExchangeType_`
        :param routing_key: routing key
        :type routing_key: str
        :param arguments: additional arguments (will be passed to `pika`)
        :type arguments: dict or NoneType
        :param timeout: execution timeout
        :type timeout: int
        :raises tornado.gen.TimeoutError: when the binding timeout period has elapsed.
        :rtype: :class:`tornado.concurrent.Future`
        """
        kwargs = dict(arguments=arguments, timeout=timeout)

        result = yield super(RobustQueue, self).bind(exchange=exchange, routing_key=routing_key, **kwargs)

        self._bindings[(exchange, routing_key)] = kwargs

        raise gen.Return(result)

    @gen.coroutine
    def unbind(self, exchange, routing_key, arguments=None, timeout=None):
        """
        :param exchange:  The exchange to unbind
        :type exchange: :class:`ExchangeType_`
        :param routing_key: The routing key
        :type routing_key: str
        :param arguments:
        :param timeout:
        :return:
        """
        result = yield super(RobustQueue, self).unbind(exchange, routing_key, arguments, timeout)
        self._bindings.pop((exchange, routing_key), None)

        raise gen.Return(result)

    @tools.coroutine
    def consume(self, callback, no_ack=False, exclusive=False, arguments=None, consumer_tag=None, timeout=None):
        # pylint: disable=too-many-arguments
        """ Start to consuming the :class:`Queue`.

        :param callback: Consuming callback. Could be a coroutine.
        :type callback: :class:`types.FunctionType`
        :param no_ack: if :class:`True` you don't need to call :func:`topika.message.IncomingMessage.ack`
        :type no_ack: bool
        :param exclusive: Makes this queue exclusive. Exclusive queues may only be accessed by the current
                          connection, and are deleted when that connection closes. Passive declaration of an
                          exclusive queue by other connections are not allowed.
        :type exclusive: bool
        :param arguments: extended arguments for pika
        :type arguments: Optiona[dict]
        :param consumer_tag: optional consumer tag
        :param timeout: :class:`tornado.gen.TimeoutError` will be raises when the
                        Future was not finished after this time.

        :raises tornado.gen.TimeoutError: when the consuming timeout period has elapsed.
        :rtype: class:`Generator[Any, None, ConsumerTag]`
        """
        kwargs = dict(callback=callback, no_ack=no_ack, exclusive=exclusive, arguments=arguments)

        consumer_tag = yield super(RobustQueue, self).consume(consumer_tag=consumer_tag, **kwargs)

        self._consumers[consumer_tag] = kwargs

        raise gen.Return(consumer_tag)

    @gen.coroutine
    def cancel(self, consumer_tag, timeout=None):
        result = yield super(RobustQueue, self).cancel(consumer_tag, timeout)
        self._consumers.pop(consumer_tag, None)
        raise gen.Return(result)


__all__ = ('RobustQueue',)



from __future__ import absolute_import
import collections
import logging
import pika
import pika.exceptions
from tornado import gen, locks
from typing import Callable, Any, Generator, Union

from .common import BaseChannel
from . import compat
from .compat import Awaitable
from . import common
from . import exceptions
from . import exchange
from . import message
from . import queue
from . import tools
from . import transaction

LOGGER = logging.getLogger(__name__)

FunctionOrCoroutine = Union[Callable[[message.IncomingMessage], Any], Awaitable[message.IncomingMessage]]


class Channel(BaseChannel):
    """ Channel abstraction """

    QUEUE_CLASS = queue.Queue
    EXCHANGE_CLASS = exchange.Exchange

    __slots__ = ('_connection', '__closing', '_confirmations', '_delivery_tag', 'loop', '_futures', '_channel',
                 '_on_return_callbacks', 'default_exchange', '_write_lock', '_channel_number', '_publisher_confirms',
                 '_on_return_raises')

    def __init__(self,
                 connection,
                 loop,
                 future_store,
                 channel_number=None,
                 publisher_confirms=True,
                 on_return_raises=False):
        """
        Create a new instance of the Channel.  Don't call this directly, this should
        be constructed by the connection.

        :type connection: :class:`pika.adapters.tornado_connection.TornadoConnection`
        :type loop: :class:`tornado.ioloop.IOLoop`
        :param future_store: The future store to use
        :type future_store: :class:`topika.common.FutureStore`
        :type channel_number: int
        :type publisher_confirms: bool
        :type on_return_raises: bool
        """
        super(Channel, self).__init__(loop, future_store.create_child())

        self._channel = None  # type: pika.channel.Channel
        self._connection = connection
        self._confirmations = {}
        self._on_return_callbacks = []
        self._delivery_tag = 0
        self._write_lock = locks.Lock()
        self._channel_number = channel_number
        self._publisher_confirms = publisher_confirms

        if not publisher_confirms and on_return_raises:
            raise RuntimeError('on_return_raises must be uses with publisher confirms')

        self._on_return_raises = on_return_raises

        self.default_exchange = self.EXCHANGE_CLASS(
            loop=self.loop,
            future_store=self._futures.create_child(),
            channel=self._channel,
            publish_method=self._publish,
            name='',
            type=exchange.ExchangeType.DIRECT,
            passive=None,
            durable=None,
            auto_delete=None,
            internal=None,
            arguments=None)

    @property
    def _channel_maker(self):
        assert self._connection
        return self._connection._connection.channel

    @property
    def number(self):
        return self._channel.channel_number

    def __str__(self):
        return "{0}".format(self.number if self._channel else "Not initialized channel")

    def __repr__(self):
        return '<%s "%s#%s">' % (self.__class__.__name__, self._connection, self)

    def _on_channel_close(self, channel, reason):
        """
        :type channel: :class:`pika.channel.Channel`
        :type reason: Exception
        :return:
        """
        # Check if we initiated the channel close
        if isinstance(reason, pika.exceptions.ChannelClosedByClient):
            self._closing.set_result(reason.reply_code)
            log_method = LOGGER.debug
        else:
            self._closing.set_exception(reason)
            log_method = LOGGER.error

        log_method("Channel %r closed: %s", channel, reason)

        self._futures.reject_all(reason)
        return reason

    def _on_return(self, channel, method, properties, body):
        """Called as the result of Basic.Return from broker in
        publisher-acknowledgements mode. Saves the info as a ReturnedMessage
        instance in self._puback_return.

        :param pika.Channel channel: our self._impl channel
        :param pika.spec.Basic.Return method:
        :param pika.spec.BasicProperties properties: message properties
        :param body: returned message body; empty string if no body
        :type body: str, unicode

        """
        msg = message.ReturnedMessage(channel=channel, body=body, envelope=method, properties=properties)

        for callback in self._on_return_callbacks:
            tools.create_task(callback(msg))

    def add_close_callback(self, callback):
        """
        :type callback: :class:`FunctionType`
        """
        self._closing.add_done_callback(lambda r: callback(r))

    def remove_close_callback(self, callback):
        """
        :type callback: :class:`FunctionType`
        """
        self._closing.remove_done_callback(callback)

    @property
    @gen.coroutine
    def closing(self):
        """ Return future which will be finished after channel close. """
        raise gen.Return((yield self._closing))

    def add_on_return_callback(self, callback):
        """
        :param callback: :class:`FunctionOrCoroutine`
        """
        self._on_return_callbacks.append(gen.coroutine(callback))

    @gen.coroutine
    def _create_channel(self, timeout=None):
        future = self._create_future(timeout=timeout)

        self._channel_maker(channel_number=self._channel_number, on_open_callback=future.set_result)

        channel = yield future  # type: pika.channel.Channel
        if self._publisher_confirms:
            channel.confirm_delivery(self._on_delivery_confirmation)

            if self._on_return_raises:
                channel.add_on_return_callback(self._on_return_delivery)

        channel.add_on_close_callback(self._on_channel_close)
        channel.add_on_return_callback(self._on_return)

        raise gen.Return(channel)

    @gen.coroutine
    def initialize(self, timeout=None):
        with (yield self._write_lock.acquire()):
            if self._closing.done():
                raise RuntimeError("Can't initialize closed channel")

            self._channel = yield self._create_channel(timeout)
            self._delivery_tag = 0

    def _on_return_delivery(self, channel, method_frame, properties, body):
        f = self._confirmations.pop(int(properties.headers.get('delivery-tag')))
        f.set_exception(exceptions.UnroutableError([body]))

    def _on_delivery_confirmation(self, method_frame):
        future = self._confirmations.pop(method_frame.method.delivery_tag, None)

        if not future:
            LOGGER.info("Unknown delivery tag %d for message confirmation \"%s\"", method_frame.method.delivery_tag,
                        method_frame.method.NAME)
            return

        try:
            confirmation_type = common.ConfirmationTypes(method_frame.method.NAME.split('.')[1].lower())

            if confirmation_type == common.ConfirmationTypes.ACK:
                future.set_result(True)
            elif confirmation_type == common.ConfirmationTypes.NACK:
                future.set_exception(exceptions.NackError(method_frame))
        except ValueError:
            future.set_exception(RuntimeError('Unknown method frame', method_frame))
        except Exception as e:
            future.set_exception(e)

    @BaseChannel._ensure_channel_is_open
    @gen.coroutine
    def declare_exchange(self,
                         name,
                         type=exchange.ExchangeType.DIRECT,
                         durable=None,
                         auto_delete=False,
                         internal=False,
                         passive=False,
                         arguments=None,
                         timeout=None):
        """
        :type name: str
        :type type: ExchangeType
        :type durable: Optional[bool]
        :type auto_delete: Optional[bool]
        :type internal: Optional[bool]
        :type passive: Optional[bool]
        :type arguments: dict or NoneType
        :type timeout: int
        :rtype: :class:`Generator[Any, None, exchange.Exchange]`
        """

        with (yield self._write_lock.acquire()):
            if auto_delete and durable is None:
                durable = False

            exchange = self.EXCHANGE_CLASS(
                loop=self.loop,
                future_store=self._futures.create_child(),
                channel=self._channel,
                publish_method=self._publish,
                name=name,
                type=type,
                passive=passive,
                durable=durable,
                auto_delete=auto_delete,
                internal=internal,
                arguments=arguments)

            yield exchange.declare(timeout=timeout)

            LOGGER.debug("Exchange declared %r", exchange)

            raise gen.Return(exchange)

    @BaseChannel._ensure_channel_is_open
    @gen.coroutine
    def _publish(self, queue_name, routing_key, body, properties, mandatory):
        """
        :type properties: :class:`pika.BasicProperties`
        """
        with (yield self._write_lock.acquire()):
            while self._connection.is_closed:
                LOGGER.debug("Can't publish message because connection is inactive")
                yield gen.sleep(1)

            publish_future = self._create_future()
            self._delivery_tag += 1

            if self._on_return_raises:
                properties.headers = properties.headers or {}
                properties.headers['delivery-tag'] = str(self._delivery_tag)

            try:
                self._channel.basic_publish(queue_name, routing_key, body, properties, mandatory)
            except (AttributeError, RuntimeError) as exc:
                LOGGER.exception("Failed to send data to client (connection unexpectedly closed)")
                self._on_channel_close(self._channel, exc)
                self._connection._connection.close(reply_code=500, reply_text="Incorrect state")
            else:
                if self._publisher_confirms:
                    self._confirmations[self._delivery_tag] = publish_future
                else:
                    publish_future.set_result(None)

            result = yield publish_future
            raise gen.Return(result)

    @BaseChannel._ensure_channel_is_open
    @gen.coroutine
    def declare_queue(self,
                      name=None,
                      durable=None,
                      exclusive=False,
                      passive=False,
                      auto_delete=False,
                      arguments=None,
                      timeout=None):
        """
        :param name: queue name
        :type name: str
        :param durable: Durability (queue survive broker restart)
        :type durable: bool
        :param exclusive: Makes this queue exclusive. Exclusive queues may only be \
        accessed by the current connection, and are deleted when that connection \
        closes. Passive declaration of an exclusive queue by other connections are not allowed.
        :type exclusive: bool
        :param passive: Only check to see if the queue exists.
        :type passive: bool
        :param auto_delete: Delete queue when channel will be closed.
        :type auto_delete: bool
        :param arguments: pika additional arguments
        :type arguments: Optional[dict]
        :param timeout: execution timeout
        :type timeout: int
        :rtype: :class:`topika.Queue`
        """

        with (yield self._write_lock.acquire()):
            if auto_delete and durable is None:
                durable = False

            queue = self.QUEUE_CLASS(self.loop, self._futures.create_child(), self._channel, name, durable, exclusive,
                                     auto_delete, arguments)

            yield queue.declare(timeout, passive=passive)
            raise gen.Return(queue)

    @gen.coroutine
    def close(self):
        if not self._channel:
            return

        with (yield self._write_lock.acquire()):
            self._channel.close()
            yield self.closing
            self._channel = None

    @BaseChannel._ensure_channel_is_open
    @gen.coroutine
    def set_qos(self, prefetch_count=0, prefetch_size=0, global_qos=False, timeout=None):
        """
        :type prefetch_count: int
        :type prefetch_size: int
        :type global_qos: bool
        :type timeout: int
        """

        with (yield self._write_lock.acquire()):
            f = self._create_future(timeout=timeout)

            self._channel.basic_qos(
                prefetch_size=prefetch_size,
                prefetch_count=prefetch_count,
                global_qos=global_qos,
                callback=f.set_result,
            )

            raise gen.Return((yield f))

    @BaseChannel._ensure_channel_is_open
    @gen.coroutine
    def queue_delete(self, queue_name, timeout=None, if_unused=False, if_empty=False):
        """
        :type queue_name: str
        :type timeout: int
        :type if_unused: bool
        :type if_empty: bool
        """
        with (yield self._write_lock.acquire()):
            f = self._create_future(timeout=timeout)

            self._channel.queue_delete(
                callback=f.set_result,
                queue=queue_name,
                if_unused=if_unused,
                if_empty=if_empty,
            )

            raise gen.Return((yield f))

    @BaseChannel._ensure_channel_is_open
    @gen.coroutine
    def exchange_delete(self, exchange_name, timeout=None, if_unused=False):
        """
        :type exchange_name: str
        :type timeout: int
        :type if_unused:bool
        """
        with (yield self._write_lock.acquire()):
            f = self._create_future(timeout=timeout)

            self._channel.exchange_delete(exchange=exchange_name, if_unused=if_unused, callback=f.set_result)

            raise gen.Return((yield f))

    def transaction(self):
        """
        :rtype: :class:`topika.Transaction`
        """
        if self._publisher_confirms:
            raise RuntimeError("Cannot create transaction when publisher " "confirms are enabled")

        tx = transaction.Transaction(self._channel, self._futures.create_child())

        self.add_close_callback(tx.on_close_callback)

        tx.closing.add_done_callback(lambda _: self.remove_close_callback(tx.on_close_callback))

        return tx


__all__ = ('Channel',)



from __future__ import absolute_import
import functools
import pika.exceptions
import tornado.concurrent
from tornado import gen

__all__ = 'wait', 'create_future', 'create_task', 'iscoroutinepartial'


def iscoroutinepartial(coro):
    """
    Function returns True if function it's a partial instance of coroutine. See additional information here_.

    :param coro: Function
    :return: bool

    .. _here: https://goo.gl/C0S4sQ

    """

    while True:
        parent = coro

        coro = getattr(parent, 'func', None)

        if coro is None:
            break

    return gen.is_coroutine_function(parent)


def create_future(loop):
    """ Helper for `create a new future`_ with backward compatibility for Python 3.4

    .. _create a new future: https://goo.gl/YrzGQ6
    """

    try:
        return loop.create_future()
    except AttributeError:
        # Compatibility with older tornado
        return tornado.concurrent.Future()


def create_task(yielded):
    """ Helper for `create a new Task`_ with backward compatibility for Python 3.4

    .. _create a new Task: https://goo.gl/g4pMV9
    """

    return gen.convert_yielded(yielded)


@gen.coroutine
def wait(tasks):
    """
    Simple helper for gathering all passed :class:`Task`s.

    :param tasks: list of the :class:`asyncio.Task`s
    :param _loop: Event loop (:func:`tornado.ioloop.IOLoop.current()` when :class:`None`)
    :return: :class:`tuple` of results
    """
    raise gen.Return((yield gen.multi(tasks)))


def ensure_connection_exception(exception_or_message):
    """
    If passed an exception this will be returned.  Otherwise it is assumed
    a string is passed giving the reason for the connection error

    :param exception_or_message:
    :return:
    """
    if isinstance(exception_or_message, Exception):
        return exception_or_message

    # We got a string message
    return pika.exceptions.AMQPConnectionError(exception_or_message)


# Get rid of the stupid default replace_callback in tornado coroutine
coroutine = functools.partial(gen._make_coroutine_wrapper, replace_callback=False)



from __future__ import absolute_import
from functools import wraps
from logging import getLogger
from typing import Callable, Generator, Any
import pika.channel
from pika.exceptions import ChannelClosed
from tornado import gen

from . import compat
from .exceptions import ProbableAuthenticationError
from .connection import Connection, connect
from .robust_channel import RobustChannel

log = getLogger(__name__)


def _ensure_connection(func):

    @wraps(func)
    def wrap(self, *args, **kwargs):
        if self.is_closed:
            raise RuntimeError("Connection closed")

        return func(self, *args, **kwargs)

    return wrap


class RobustConnection(Connection):
    """ Robust connection """

    DEFAULT_RECONNECT_INTERVAL = 1
    CHANNEL_CLASS = RobustChannel

    def __init__(self,
                 host='localhost',
                 port=5672,
                 login='guest',
                 password='guest',
                 virtual_host='/',
                 loop=None,
                 **kwargs):
        """
        :type host: str
        :type port: int
        :type login: str
        :type password: str
        :type virtual_host: str
        :type loop: :class:`tornado.ioloop.IOLoop`
        :type kwargs: dict
        """

        self.reconnect_interval = kwargs.pop('reconnect_interval', self.DEFAULT_RECONNECT_INTERVAL)

        super(RobustConnection, self).__init__(
            host=host, port=port, login=login, password=password, virtual_host=virtual_host, loop=loop, **kwargs)

        self._closed = False
        self._on_connection_lost_callbacks = []
        self._on_reconnect_callbacks = []
        self._on_close_callbacks = []

    def add_connection_lost_callback(self, callback):
        """ Add callback which will be called after connection was lost.

        :type callback: :class:`Callable[[], None]`
        :return: None
        """

        self._on_connection_lost_callbacks.append(lambda c: callback(c))

    def add_reconnect_callback(self, callback):
        """ Add callback which will be called after reconnect.

        :type callback: :class:`Callable[[], None]`
        :return: None
        """

        self._on_reconnect_callbacks.append(lambda c: callback(c))

    def add_close_callback(self, callback):
        """ Add callback which will be called after connection will be closed.

        :type callback: :class:`Callable[[], None]`
        :return: None
        """

        self._on_close_callbacks.append(lambda c: callback(c))

    def _on_connection_lost(self, future, connection, reason):
        """
        :type future: :class:`tornado.concurrent.Future`
        :type connection: :class:`pika.adapters.tornado_connection.TornadoConnection`
        :type reason: Exception
        """
        for callback in self._on_connection_lost_callbacks:
            callback(self)

        if self._closed:
            super(RobustConnection, self)._on_connection_lost(future, connection, reason)

        if isinstance(reason, ProbableAuthenticationError):
            log.error("Authentication error: %s", reason)

        if isinstance(reason, compat.ConnectionRefusedError):
            log.error("Connection refused: %s", reason)

        if not future.done():
            future.set_result(None)

        self.loop.call_later(self.reconnect_interval, self.connect)

    def _channel_cleanup(self, channel):
        """
        :type channel: :class:`pika.channel.Channel`
        """
        pika_channel = self._channels[channel.channel_number]  # type: RobustChannel
        pika_channel._futures.reject_all(ChannelClosed)

        if pika_channel._closed:
            self._channels.pop(channel.channel_number)  # type: RobustChannel

    def _on_channel_error(self, channel):
        """
        :type channel: :class:`pika.channel.Channel`
        """
        log.error("Channel closed: %s. Will attempt to reconnect", channel)
        channel.connection.close(reply_code=500, reply_text="Channel canceled")

    def _on_channel_cancel(self, channel):
        """
        :type channel: :class:`pika.channel.Channel`
        """
        log.debug("Channel canceled: %s", channel)
        self._on_channel_error(channel)

    @gen.coroutine
    def connect(self):
        result = yield super(RobustConnection, self).connect()

        while self._connection is None:
            yield gen.sleep(self.reconnect_interval)
            result = yield super(RobustConnection, self).connect()

        for number, channel in tuple(self._channels.items()):
            try:
                yield channel.on_reconnect(self, number)
            except ChannelClosed:
                self._on_channel_error(channel._channel)
                return

        for callback in self._on_reconnect_callbacks:
            callback(self)

        raise gen.Return(result)

    @property
    def is_closed(self):
        """ Is this connection is closed """

        return self._closed or super(RobustConnection, self).is_closed

    def close(self):
        """
        Close AMQP connection

        :rtype: :class:`tornado.concurrent.Future`
        """
        self._closed = True

        try:
            for callback in self._on_close_callbacks:
                callback(self)
        finally:
            return super(RobustConnection, self).close()


@gen.coroutine
def connect_robust(url=None,
                   host='localhost',
                   port=5672,
                   login='guest',
                   password='guest',
                   virtualhost='/',
                   loop=None,
                   ssl_options=None,
                   connection_class=RobustConnection,
                   **kwargs):
    """ Make robust connection to the broker.

    That means that connection state will be restored after reconnect.
    After connection has been established the channels, the queues and the
    exchanges with their bindings will be restored.

    Example:

    .. code-block:: python

        import topika
        import tornado.gen

        @tornado.gen.coroutine
        def main():
            connection = yield topika.connect_robust("amqp://guest:guest@127.0.0.1/")

    Connect to localhost with default credentials:

    .. code-block:: python

        import topika
        import tornado.gen

        @tornado.gen.coroutine
        def main():
            connection = yield topika.connect_robust()

    .. note::

        The available keys for ssl_options parameter are:
            * cert_reqs
            * certfile
            * keyfile
            * ssl_version

        For an information on what the ssl_options can be set to reference the
        `official Python documentation`_.

        .. _official Python documentation: http://docs.python.org/3/library/ssl.html

    URL string might be contain ssl parameters e.g.
    `amqps://user:password@10.0.0.1//?ca_certs=ca.pem&certfile=cert.pem&keyfile=key.pem`

    :param url: `RFC3986`_ formatted broker address. When :class:`None` \
                will be used keyword arguments.
    :type url: str
    :param host: hostname of the broker
    :type host: str
    :param port: broker port 5672 by default
    :type port: int
    :param login: username string. `'guest'` by default. Provide empty string \
                  for pika.credentials.ExternalCredentials usage.
    :type login: str
    :param password: password string. `'guest'` by default.
    :type password: str
    :param virtualhost: virtualhost parameter. `'/'` by default
    :type virtualhost: str
    :param ssl_options: A dict of values for the SSL connection.
    :type ssl_options: dict
    :param loop: Event loop (:func:`tornado.ioloop.IOLoop.current()` when :class:`None`)
    :type loop: :class:`tornado.ioloop.IOLoop`
    :param connection_class: Factory of a new connection
    :param kwargs: addition parameters which will be passed to \
                   the pika connection.
    :rtype: :class:`topika.connection.Connection`

    .. _RFC3986: https://tools.ietf.org/html/rfc3986
    .. _pika documentation: https://goo.gl/TdVuZ9

    """
    raise gen.Return((yield connect(
        url=url,
        host=host,
        port=port,
        login=login,
        password=password,
        virtualhost=virtualhost,
        loop=loop,
        connection_class=connection_class,
        ssl_options=ssl_options,
        **kwargs)))


__all__ = 'RobustConnection', 'connect_robust',



from __future__ import absolute_import
from pika.exceptions import (ProbableAuthenticationError, AMQPChannelError, AMQPConnectionError, AMQPError,
                             ChannelClosed, ChannelError, AuthenticationError, BodyTooLongError, ConnectionClosed,
                             ConsumerCancelled, DuplicateConsumerTag, IncompatibleProtocolError, InvalidChannelNumber,
                             InvalidFieldTypeException, InvalidFrameError, MethodNotImplemented, NackError, NoFreeChannels,
                             ProbableAccessDeniedError, ProtocolSyntaxError, ProtocolVersionMismatch,
                             ShortStringTooLong, UnexpectedFrameError, UnroutableError, UnsupportedAMQPFieldException)


class AMQPException(Exception):
    pass


class MessageProcessError(AMQPException):
    pass


class QueueEmpty(AMQPException):
    pass


class TransactionClosed(AMQPException):
    pass


__all__ = (
    'AMQPChannelError',
    'AMQPConnectionError',
    'AMQPError',
    'AMQPException',
    'AuthenticationError',
    'BodyTooLongError',
    'ChannelClosed',
    'ChannelError',
    'ConnectionClosed',
    'ConsumerCancelled',
    'DuplicateConsumerTag',
    'IncompatibleProtocolError',
    'InvalidChannelNumber',
    'InvalidFieldTypeException',
    'InvalidFrameError',
    'MessageProcessError',
    'MethodNotImplemented',
    'NackError',
    'NoFreeChannels',
    'ProbableAccessDeniedError',
    'ProbableAuthenticationError',
    'ProtocolSyntaxError',
    'ProtocolVersionMismatch',
    'QueueEmpty',
    'ShortStringTooLong',
    'TransactionClosed',
    'UnexpectedFrameError',
    'UnroutableError',
    'UnsupportedAMQPFieldException',
)

try:
    ConnectionError
except NameError:

    class ConnectionError(Exception):
        pass

    class ConnectionRefusedError(ConnectionError):
        pass



from __future__ import absolute_import
import logging
from functools import wraps, partial
import pika
from pika import ConnectionParameters
from pika.credentials import ExternalCredentials, PlainCredentials
from pika.spec import REPLY_SUCCESS
from pika.adapters.tornado_connection import TornadoConnection
import pika.exceptions
import tornado.ioloop
from tornado import gen, ioloop, locks
from tornado.concurrent import Future
from tornado.gen import coroutine, Return
from six.moves.urllib.parse import urlparse

from .channel import Channel
from . import common
from . import compat
from . import exceptions
from . import tools

__all__ = ('Connection', 'connect')

LOGGER = logging.getLogger(__name__)


class Connection(object):
    """ Connection abstraction """

    __slots__ = ('loop', '__closing', '_connection', 'future_store', '__sender_lock', '_io_loop',
                 '__connection_parameters', '__credentials', '__write_lock', '_channels', '__close_started')

    CHANNEL_CLASS = Channel

    def __init__(self,
                 host='localhost',
                 port=5672,
                 login='guest',
                 password='guest',
                 virtual_host='/',
                 loop=None,
                 **kwargs):

        self.loop = loop if loop else ioloop.IOLoop.current()
        self.future_store = common.FutureStore(loop=self.loop)

        self.__credentials = PlainCredentials(login, password) if login else ExternalCredentials()

        self.__connection_parameters = ConnectionParameters(
            host=host, port=port, credentials=self.__credentials, virtual_host=virtual_host, **kwargs)

        self._channels = dict()
        self._connection = None
        self.__closing = None
        self.__close_started = False
        self.__write_lock = locks.Lock()

    def __str__(self):
        return 'amqp://{credentials}{host}:{port}/{vhost}'.format(
            credentials="{0.username}:********@".format(self.__credentials) if isinstance(
                self.__credentials, PlainCredentials) else '',
            host=self.__connection_parameters.host,
            port=self.__connection_parameters.port,
            vhost=self.__connection_parameters.virtual_host,
        )

    def __repr__(self):
        cls_name = self.__class__.__name__
        return '<{0}: "{1}">'.format(cls_name, str(self))

    def add_close_callback(self, callback):
        """ Add callback which will be called after connection will be closed.

        :class:`tornado.concurrent..Future` will be passed as a first argument.

        Example:

        .. code-block:: python

            import topika
            from tornado import gen

            @gen.coroutine
            def main():
                connection = yield topika.connect(
                    "amqp://guest:guest@127.0.0.1/"
                )
                connection.add_close_callback(print)
                yield connection.close()
                # <Future finished result='Normal shutdown'>


        :type callback: Callable[[], None]
        :return: None
        """
        self._closing.add_done_callback(callback)

    @property
    def is_closed(self):
        """ Is this connection closed """

        if not self._connection:
            return True

        if self._closing.done():
            return True

        return False

    @property
    def _closing(self):
        self._ensure_cosing_future()
        return self.__closing

    def _ensure_cosing_future(self, force=False):
        if self.__closing is None or force:
            self.__closing = self.future_store.create_future()

    @property
    @gen.coroutine
    def closing(self):
        """ Return coroutine which will be finished after connection close.

        Example:

        .. code-block:: python

            import topika
            import tornado.gen

            @tornado.gen.coroutine
            def async_close(connection):
                yield tornado.gen.sleep(2)
                yield connection.close()

            @tornado.gen.coroutine
            def main(loop):
                connection = yield topika.connect(
                    "amqp://guest:guest@127.0.0.1/"
                )
                topika.create_task(async_close(connection))

                yield connection.closing

        """
        raise gen.Return((yield self._closing))

    def __del__(self):
        with compat.suppress():
            if not self.is_closed:
                self.close()

    @gen.coroutine
    def connect(self):
        """ Connect to AMQP server. This method should be called after :func:`topika.connection.Connection.__init__`

        .. note::
            This method is called by :func:`connect`. You shouldn't call it explicitly.

        :rtype: :class:`pika.adapters.tornado_connection.TornadoConnection`
        """

        if self.__closing and self.__closing.done():
            raise RuntimeError("Invalid connection state")

        with (yield self.__write_lock.acquire()):
            self._connection = None

            LOGGER.debug("Creating a new AMQP connection: %s", self)

            connect_future = tools.create_future(loop=self.loop)

            connection = TornadoConnection(
                parameters=self.__connection_parameters,
                custom_ioloop=self.loop,
                on_open_callback=connect_future.set_result,
                on_close_callback=partial(self._on_connection_lost, connect_future),
                on_open_error_callback=partial(self._on_connection_refused, connect_future),
            )

            connection.channel_cleanup_callback = self._channel_cleanup
            connection.channel_cancel_callback = self._on_channel_cancel

            result = yield connect_future

            LOGGER.debug("Connection ready: %r", self)

            self._connection = connection
            raise gen.Return(result)

    @gen.coroutine
    def channel(self, channel_number=None, publisher_confirms=True, on_return_raises=False):
        """ Coroutine which returns new instance of :class:`Channel`.

        Example:

        .. code-block:: python
            from tornado import gen
            import topika

            @gen.coroutine
            def main(loop):
                connection = yield topika.connect(
                    "amqp://guest:guest@127.0.0.1/"
                )

                channel1 = yield connection.channel()
                yield channel1.close()

                # Creates channel with specific channel number
                channel42 = yield connection.channel(42)
                yield channel42.close()

                # For working with transactions
                channel_no_confirms = connection.channel(publisher_confirms=True)
                yield channel_no_confirms.close()

        :param channel_number: specify the channel number explicit
        :type channel_number: int
        :param publisher_confirms:
            if `True` the :func:`aio_pika.Exchange.publish` method will be return
            :class:`bool` after publish is complete. Otherwise the
            :func:`aio_pika.Exchange.publish` method will be return :class:`None`
        :type publisher_confirms: bool
        :param on_return_raises:
            raise an :class:`topika.exceptions.UnroutableError`
            when mandatory message will be returned
        :rtype: :class:`Generator[Any, None, Channel]`
        """
        with (yield self.__write_lock.acquire()):
            LOGGER.debug("Creating AMQP channel for conneciton: %r", self)

            channel = self.CHANNEL_CLASS(
                self,
                self.loop,
                self.future_store,
                channel_number=channel_number,
                publisher_confirms=publisher_confirms,
                on_return_raises=on_return_raises)
            yield channel.initialize()

            LOGGER.debug("Channel created: %r", channel)

            self._channels[channel.number] = channel

            raise gen.Return(channel)

    def close(self):
        """
        Close AMQP connection

        This method is idempotent and may be called multiple times without conflict
        """
        LOGGER.debug("Closing AMQP connection")

        if self.__close_started:
            return self.__closing

        @gen.coroutine
        def inner():
            if self._connection:
                self._connection.close()
            yield self.closing

        self.__close_started = True
        return tools.create_task(inner())

    #
    # Connections state properties
    #

    @property
    def is_open(self):
        """
        Returns a boolean reporting the current connection state.
        """
        return self._connection.is_open

    def _channel_cleanup(self, channel):
        """
        :type channel: :class:`pika.channel.Channel`
        """
        ch = self._channels.pop(channel.channel_number)  # type: Channel
        ch._futures.reject_all(exceptions.ChannelClosed)

    def _on_connection_refused(self, future, connection, reason):
        """
        :type future: :class:`tornado.concurrent.Future`
        :type connection: :class:`pika.adapters.tornado_connection.TornadoConnection`
        :type reason: Exception
        """
        self._on_connection_lost(future, connection, reason)

    def _on_connection_lost(self, future, connection, reason):
        """
        :type future: :class:`tornado.concurrent.Future`
        :type connection: :class:`pika.adapters.tornado_connection.TornadoConnection`
        :type reason: Exception
        """
        if self.__closing and self.__closing.done():
            return

        if isinstance(reason, pika.exceptions.ConnectionClosedByClient) and \
                reason.reply_code == REPLY_SUCCESS:
            return self.__closing.set_result(reason)

        self.future_store.reject_all(reason)

        if future.done():
            return

        future.set_exception(reason)

    def _on_channel_cancel(self, channel):
        """
        :type channel: :class:`pika.channel.Channel`
        """
        ch = self._channels.pop(channel.channel_number)  # type: Channel
        ch._futures.reject_all(exceptions.ChannelClosed)

    #
    # Properties that reflect server capabilities for the current connection
    #

    @property
    def basic_nack_supported(self):
        """Specifies if the server supports basic.nack on the active connection.

        :rtype: bool

        """
        return self._connection.basic_nack

    @property
    def consumer_cancel_notify_supported(self):
        """Specifies if the server supports consumer cancel notification on the
        active connection.

        :rtype: bool

        """
        return self._connection.consumer_cancel_notify

    @property
    def exchange_exchange_bindings_supported(self):
        """Specifies if the active connection supports exchange to exchange
        bindings.

        :rtype: bool

        """
        return self._connection.exchange_exchange_bindings

    @property
    def publisher_confirms_supported(self):
        """Specifies if the active connection can use publisher confirmations.

        :rtype: bool

        """
        return self._connection.publisher_confirms

    # Legacy property names for backward compatibility
    basic_nack = basic_nack_supported
    consumer_cancel_notify = consumer_cancel_notify_supported
    exchange_exchange_bindings = exchange_exchange_bindings_supported
    publisher_confirms = publisher_confirms_supported

    def ensure_connected(self):
        if self.is_closed:
            raise RuntimeError("Connection closed")


@gen.coroutine
def connect(url=None,
            host='localhost',
            port=5672,
            login='guest',
            password='guest',
            virtualhost='/',
            loop=None,
            ssl_options=None,
            connection_class=Connection,
            **kwargs):
    """ Make connection to the broker.

    Example:

    .. code-block:: python

        import topika
        import tornado.gen

        @tornado.gen.coroutine
        def main():
            connection = yield topika.connect(
                "amqp://guest:guest@127.0.0.1/"
            )

    Connect to localhost with default credentials:

    .. code-block:: python

        import topika
        import tornado.gen

        @tornado.gen.coroutine
        def main():
            connection = yield topika.connect()

    .. note::

        The available keys for ssl_options parameter are:
            * cert_reqs
            * certfile
            * keyfile
            * ssl_version

        For an information on what the ssl_options can be set to reference the
        `official Python documentation`_.

        .. _official Python documentation: http://docs.python.org/3/library/ssl.html

    URL string might be contain ssl parameters e.g.
    `amqps://user:password@10.0.0.1//?ca_certs=ca.pem&certfile=cert.pem&keyfile=key.pem`

    :param url: `RFC3986`_ formatted broker address. When :class:`None` \
                will be used keyword arguments.
    :type url: str
    :param host: hostname of the broker
    :type host: str
    :param port: broker port 5672 by default
    :type port: int
    :param login: username string. `'guest'` by default. Provide empty string \
                  for pika.credentials.ExternalCredentials usage.
    :type login: str
    :param password: password string. `'guest'` by default.
    :type password: str
    :param virtualhost: virtualhost parameter. `'/'` by default
    :type virtualhost: str
    :param ssl_options: A dict of values for the SSL connection.
    :type ssl_options: dict
    :param loop: Event loop (:func:`tornado.ioloop.IOLoop.current()` when :class:`None`)
    :type loop: :class:`tornado.ioloop.IOLoop`
    :param connection_class: Factory of a new connection
    :param kwargs: addition parameters which will be passed to \
                   the pika connection.
    :rtype: :class:`topika.connection.Connection`

    .. _RFC3986: https://tools.ietf.org/html/rfc3986
    .. _pika documentation: https://goo.gl/TdVuZ9

    """
    if url:
        url = urlparse(str(url))
        host = url.hostname or host
        port = url.port or port
        login = url.username or login
        password = url.password or password
        virtualhost = url.path[1:] if len(url.path) > 1 else virtualhost

        ssl_keys = (
            'ca_certs',
            'cert_reqs',
            'certfile',
            'keyfile',
            'ssl_version',
        )

        for key in ssl_keys:
            if key not in url.query:
                continue

            ssl_options[key] = url.query[key]

    connection = connection_class(
        host=host,
        port=port,
        login=login,
        password=password,
        virtual_host=virtualhost,
        loop=loop,
        ssl_options=ssl_options,
        **kwargs)

    yield connection.connect()
    raise gen.Return(connection)
